#
# kickstart.py - SNAKE kickstart templates
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#


import os
import sys
import shlex
import __builtin__
import pykickstart.parser
from pykickstart.version import returnClassForVersion, stringToVersion
from pykickstart.errors import KickstartVersionError
from pykickstart.base import BaseHandler
from snake.config import config

header = '''\
# kickstart automatically generated by SNAKE.
# Please don't modify the following two lines.
# Description: %s
# Version: %s
'''

header_wargs = header + '''\
# KernelArgs: %s
'''

class Kickstart(object):
    '''A SNAKE kickstart object. Takes an optional argument that selects the
    version of the kickstart language to use. This argument may be a string
    (e.g. 'F7', 'RHEL4') or one of the constants from pykickstart.parser.

    If unspecified, the 'devel' handler class will be used. This is the
    default handler that pykickstart.parser.returnClassForVersion will return
    if called with no argument.

    This class has the following instance variables:
    description: A description of the kickstart object
    version:     The (numeric) version of the kickstart language being used
    kernel_args: Boot arguments associated with this kickstart (optional)
    handler:     A handler for the kickstart version specified
    parser:      A kickstart parser (uses the aforementioned handler)
    '''
    def __init__(self,version='devel'):
        self.description='Unmodified Kickstart object'
        self.setversion(version)
        self.kernel_args = ""

    def setversion(self,version):
        '''Set this kickstart object to use the kickstart version specified.
        Note that this will reset the handler and parser attributes.'''
        if type(version) == type(str()):
            self.version = pykickstart.parser.stringToVersion(version)
        else:
            self.version = version
        handlerclass = pykickstart.parser.returnClassForVersion(self.version)
        # Create handler instance
        self.handler = handlerclass()
        # Create parser instance that uses the given handler
        self.parser = pykickstart.parser.KickstartParser(self.handler)

    def read(self,ksfile,reset=True, ksmeta={}):
        '''Read in the given kickstart file. If reset is True, the handler
        and parser will be re-initialized based on the contents of the file.'''
        f = open(ksfile)
        self.readfp(f,reset,ksmeta)
        f.close()

    def readfp(self,fp,reset=True, ksmeta={}):
        '''Read kickstart data from the given file-like object.
        (fp must support read())
        Set the 'reset' argument to False if you don't want to re-initialize
        the parser, handler, etc.'''
        # FIXME: need to ignore deprecation warnings etc.
        self.readbuf(fp.read(),reset,ksmeta)

    def readbuf(self,buf,reset=True, ksmeta={}):
        '''Read kickstart data from the given string buffer.
        Set the 'reset' argument to False if you don't want to re-initialize
        the parser, handler, etc.'''

        # TODO - ksmeta is only honored by the PyKickstart object ... perhaps
        # ksmeta variable substitution could be honored

        data = list()
        for l in buf.split("\n"):
            data.append(l)
            if l.startswith("# Description: "):
                (k,v) = l.split(":",1)
                self.description = v.strip()
            elif l.startswith("# KernelArgs: "):
                (k,v) = l.split(":",1)
                self.kernel_args = v.strip()
            elif reset and l.startswith("# Version: "):
                (k,v) = l.split(":",1)
                self.version = int(v.strip())
        if reset:
            self.setversion(self.version)
        self.parser.errorsAreFatal = False
        self.parser.followIncludes = False
        self.parser.readKickstartFromString('\n'.join(data))

    def handlecommand(self,cmdstring):
        '''Given a kickstart command (in a string), pass that to the kickstart
        parser and modify the kickstart object accordingly.'''
        self.parser.handleCommand(0,shlex.split(cmdstring))

    def write(self,ksfile,handler=None):
        '''Write a valid kickstart file out to the given filename.'''
        f = open(ksfile,'w')
        self.writefp(f,handler)
        f.close()

    def writefp(self,fp,handler=None):
        '''Write valid kickstart data out to the given file-like object.'''
        fp.write(self.__str__(handler))

    # This needs to know what vintage of kickstart to write out. If handler is
    # left as None, we use our default handler.
    def __str__(self,handler=None):
        if self.kernel_args:
            h = header_wargs % (self.description, self.version, self.kernel_args)
        else:
            h = header % (self.description, self.version)
        if handler:
            print "FIXME: can't actually convert between handlers yet"
            handler = self.handler
        else:
            handler = self.handler
        return h + '\n' + str(handler)

def __load_module__(modname):
    import imputil
    obj = None
    try:
        found = imputil.imp.find_module(modname)
        obj = imputil.imp.load_module(modname, found[0], found[1], found[2])
    except Exception, e:
        pass
    return obj

class PyKickstart(Kickstart):

    def setversion(self,version):
        Kickstart.setversion(self,version)

        # HACK .. pykickstart isn't using the BaseHandler.version for %{pre,post,packages}
        pykickstart.parser.ver = self.version

        try:
            SnakeHandler.__bases__ = (returnClassForVersion(self.version),)
        except KickstartVersionError:
            # FIXME - raise error?
            SnakeHandler.__bases__ = (returnClassForVersion(),) # gives you the default

    def readbuf(self,code,reset=True):
        '''Read kickstart data from the given string buffer.
        Set the 'reset' argument to False if you don't want to re-initialize
        the parser, handler, etc.'''

        # the following will raise an exception if given a bad python file
        code = __builtin__.compile(code, "<string>", "exec")

        if not "ks" in code.co_names:
            raise Exception, "Invalid pykickstart template.  No method named 'ks' found"

        if reset:
            self.setversion(self.version)

        # FIXME ... we're just saving the python code for later use by writefp()
        # Look at imputil.py _compile() for an example
        self.pycode = code

    def writefp(self,fp,handler=None):
        '''Write python code out to the given file-like object.'''
        raise Exception, "PyKickstart 'writefp' not yet supported"
        fp.write(self.pycode)

    def read(self,ksfile,reset=True,ksmeta={}):
        '''Read in the given kickstart file. If reset is True, the handler
        and parser will be re-initialized based on the contents of the file.'''

        if not os.path.dirname(ksfile) in sys.path:
            sys.path.append(os.path.dirname(ksfile))

        (name, ext) = os.path.splitext(os.path.basename(ksfile))

        if reset:
            self.setversion(self.version)

        obj = __load_module__(name)

        if not obj or not obj.__dict__.has_key("ks"):
            raise Exception, "mal-formed kickstart template '%s.py'" % name

        # use docstring from ks function as description
        if obj.ks.__doc__:
            self.description = obj.ks.__doc__

        # load any kernel arguments this kickstart
        if obj.__dict__.has_key("kernel_args_hook"):
            self.kernel_args = obj.kernel_args_hook()

        self.obj = obj
        self.parser.followIncludes = False
        self.parser.readKickstartFromString(str(self.obj.ks(**ksmeta)))

class SnakeHandler(BaseHandler):
    '''The base class for a SNAKE kickstart object.'''
    def __init__(self):
        BaseHandler.__init__(self)

