#
# ksdb.py - like treedb.py, functions for handling a database of kickstarts
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

import sys
import os
import glob
import imputil
from snake.config import getoption
from snake.kickstart import Kickstart,PyKickstart
import snake.machinedb

import logging
log = logging.getLogger("snake")

KickstartNotFound = Exception()
KickstartAlreadyExists = Exception()

def name_to_ksfile(name):
    ksfile = None
    for ksfile in [os.path.join(getoption("templatedir"),"%s.py" % name),
                   os.path.join(getoption("ksdb_dir"),"%s.ks" % name)]:
        if os.path.exists(ksfile):
            break
    return ksfile

def name_to_kickstart(name):

    ksfile = name_to_ksfile(name)

    log.debug("Loading '%s' from '%s'" % (name, ksfile))

    # instantiate the correct object
    if ksfile.endswith(".ks"):
        k = Kickstart()
    elif ksfile.endswith(".py"):
        k = PyKickstart()

    # read the kicsktart file
    try:
        k.read(ksfile)
    except Exception, e:
        log.debug("Error loading '%s' - %s" % (ksfile,e))
        return None

    return k

def add(newfile,name=None,desc=None):
    '''Add the given file (or open file-like object) to the kickstart database.
    if the newfile is an open file-like object you can pass the 'name' param
    to give it a name.'''
    if hasattr(newfile,'read'):
        fp = newfile
        if not name:
            raise ValueError, "You need a name with a file-like object, yo."
    else:
        fp = open(newfile)
        if not name:
            (name,ext) = os.path.splitext(os.path.basename(newfile))

    write_data(fp.read(),name,desc)
    return True # XML-RPC requires a return value

def write_data(data,name,desc=None):
    '''Write the given kickstart data (character buffer) to an entry in the
    ksdb named 'name'.'''
    ksdb_dir = getoption("ksdb_dir")
    template_dir = getoption("templatedir")
    k = None

    # validate kickstart - caller responsible for exception handling
    try:
        # first try to read the kickstart as if it were python cod3
        k = PyKickstart()
        k.readbuf(data)
        ksfile = os.path.join(template_dir,name+'.py')
    except Exception, e:
        # interpretting as python failed, howabout a plain txt kickstart
        k = Kickstart()
        k.readbuf(data)
        ksfile = os.path.join(ksdb_dir,name+'.ks')
        pass

    if k != None:
        if desc: k.description = desc
        out = open(ksfile, 'w')
        k.writefp(out)
        out.close()
        log.debug("Wrote kickstart %s" % ksfile)
    return True # XML-RPC requires a return value

def listkickstarts(filter={}):
    '''Return a list of kickstart names and descriptions.'''
    ksdb_dir = getoption("ksdb_dir")
    template_dir = getoption("templatedir")
    name_filter = filter.get("name","")
    desc_filter = filter.get("desc","")

    sys.path.append(ksdb_dir)

    # Generate a list of candidate files
    lst = map(lambda x: (os.path.splitext(os.path.basename(x)), x),
            glob.glob(ksdb_dir + "/*.ks") + glob.glob(template_dir + "/*.py"))

    ksdict = dict()
    # loop through '.ks' and '.py' kickstart templates
    for ((ksname,ext), fname) in lst:
        if ksname.find(name_filter) < 0:
            continue

        if ext in [".ks", ".py"]:
            k = name_to_kickstart(ksname)
        else:
            # unknown file extension, ignore
            continue

        # attempt to read the kickstart file
        try:
            if k.description.find(desc_filter) < 0:
                continue
            ksdict[ksname] = k.description
        except Exception, e:
            # Something went wrong while reading that kickstart.
            #print "failed to read %s: %s" % (fname,e)
            continue

    kslist = ksdict.items()
    kslist.sort()
    return kslist

def generate(name,kwargs={}):
    '''Generate a kickstart script, using the named template, modified by
    the given arguments.
    Accepted keyword arguments:
      commands=(list): Modify the template with the given kickstart commands.
                       The argument should be a list of command lines.
      version=(string): Generate a kickstart for the given kickstart language
                       version.
      fingerprints=(list): List of system fingerprints.  When provided, the
                       machine database will be searched for any
                       machine-specific kickstart data.
    '''
    # TODO: scripts=(list), packages=(list), tree=url|id, machine=id
    # XXX where does kernel commandline come in?

    ksversion = kwargs.get("version","devel")
    ksfile = name_to_ksfile(name)

    # try generating the python version first
    if ksfile.endswith(".py"):
        k = PyKickstart(version=ksversion)

    # okay no python template, lets just try a .ks file
    elif ksfile.endswith(".ks"):
        k = Kickstart(version=ksversion)
    else:
        raise KickstartNotFound, "Kickstart file '%s' not found" % name

    k.read(ksfile, reset=False, ksmeta=kwargs.get("ksmeta",{}))

    # process and cli provided over-rides
    if 'commands' in kwargs:
        for l in kwargs.get("commands",[]):
            k.handlecommand(l)

    # search for any machine-specific kickstart values
    if 'fingerprints' in kwargs:
        matches = snake.machinedb.listmachines(fingerprint=kwargs.get("fingerprints"))
        if len(matches) == 1:
            if matches[0].ksdata:
                for l in matches[0].ksdata.split("\n"):
                    k.handlecommand(l)
        elif len(matches) > 1:
            log.debug("Kickstart generate: found %s machines with fingerprints: %s" % (len(matches), kwargs.get("fingerprints")))

    return str(k)

def rename(old,new):
    '''Rename the named kickstart to another name'''

    oldfile = name_to_ksfile(old)

    if os.path.isfile(oldfile):

        (name,ext) = os.path.splitext(oldfile)
        newfile = os.path.join(os.path.dirname(oldfile),"%s%s" % (new, ext))

        # already a kickstart with the matching name?
        if os.path.isfile(newfile):
            raise KickstartAlreadyExists, "Kickstart file named '%s' already exists" % new

        os.rename(oldfile,newfile)
        log.debug("Kickstart renamed: %s -> %s" % (oldfile, newfile))

    return True # XML-RPC requires a return value

def describe(name,desc):
    '''Update description for the named kickstart'''
    ksfile = name_to_ksfile(name)

    if ksfile.endswith(".py"):
        raise Exception, "Describing PyKickstart templates not supported"
    elif ksfile.endswith(".ks"):
        # Read existing data (removing any description)
        fp = open(ksfile, 'r')
        data = fp.read()
        fp.close()

        # Write kickstart file using new data
        write_data(data,name,desc)
        fp.close()
    else:
        raise KickstartNotFound, "Kickstart file '%s' not found" % name
    return True # XML-RPC requires a return value

def remove(name):
    '''Remove the named kickstart from the database'''
    ksfile = name_to_ksfile(name)

    if os.path.isfile(ksfile):
        os.unlink(ksfile)
        log.debug("Removed '%s'" % ksfile)
        if ksfile.endswith(".py") and os.path.isfile(ksfile + 'c'):
            '''remove the pre-compiled python .pyc'''
            os.unlink(ksfile + 'c')
            log.debug("Removed '%sc'" % ksfile)
    else:
        raise KickstartNotFound, "Kickstart file '%s' not found" % name

    # FIXME - fail if no such 'ks' found? --- or will the traceback handle that
    return True # XML-RPC requires a return value
