#
# snake/machinedb.py - functions for manipulating the machine database
#
# Copyright 2006, 2007 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat
# trademarks that are incorporated in the source code or documentation are not
# subject to the GNU General Public License and may only be used or replicated
# with the express permission of Red Hat, Inc. 
#

import sys, os, time, tempfile
import re
import snake.machine
from snake.config import getoption
from ConfigParser import SafeConfigParser
import logging

def _loadmachines(dir=None,check=True):
    '''Loads all the machineinfo files in the given dir.  Returns a list of
    Machine objects.  '''
    if dir is None:
        dir = getoption("machinedb_dir")
    machinelist = list()
    for name in os.listdir(dir):
        f = os.path.join(dir,name,"config")
        if os.path.isfile(f):
            t = snake.machine.Machine(cfg=f)
            machinelist.append(t)
    return machinelist

def _writemachine(machine):
    '''Write the given Machine object back to the machine database'''
    # create machine directory
    cfgdir = os.path.join(getoption("machinedb_dir"), str(machine.id))
    if not os.path.isdir(cfgdir):
        os.mkdir(cfgdir)

    # create machine config file
    cfg = os.path.join(cfgdir, "config")
    fp = open(cfg,"w")
    logging.debug("Writing '%s' ...\n%s" % (cfg, dict(machine.config.items("general"))))
    machine.config.write(fp)
    fp.close()

def id_to_machine(id):
    cfg = os.path.join(getoption("machinedb_dir"),"%s" % id, "config")
    if not os.path.exists(cfg):
        return None
    try:
        m = snake.machine.Machine(cfg=cfg)
    except ConfigParser.NoSectionError:
        # File exists but there's nobody home. Bluh.
        return None
    return m

def update(id, hwinfo):
    '''Update an existing Machine with new information'''

    # validate id
    if not (isinstance(id, int) or (isinstance(id, str) and id.isdigit())):
        raise ValueError("Unexpected machine id format: %s" % type(id))

    cfgdir = os.path.join(getoption("machinedb_dir"), str(id))
    cfg = os.path.join(cfgdir, "config")

    # initialize a Machine() object
    m = snake.machine.Machine()

    # open up db record
    if os.path.isfile(cfg):
        logging.debug("Loading existing machineinfo '%s'" % cfg)
        m.parse(cfg=cfg)

    # if changing the nickname, is the new name already in use?
    if hwinfo.has_key("nickname"):
        matches = listmachines(nickname=hwinfo["nickname"])
        if matches and True in map(lambda m: m.id != id, matches):
            raise ValueError("A machine nickname '%s' already exists" % hwinfo["nickname"])

    # if changing fingerprints, are they already on file?
    elif hwinfo.has_key("fingerprints") and listmachines(fingerprint=hwinfo["fingerprints"]):
        matches = listmachines(fingerprint=hwinfo["fingerprints"])
        if matches and True in map(lambda m: m.id != id, matches):
            raise ValueError("A machine with the supplied fingerprints %s already exists" % hwinfo["fingerprints"])

    # Store user-supplied values into Machine() object
    for opt in ["id", "bootargs", "nickname", "fingerprints", "arch", "ksdata"]:
        if hwinfo.has_key(opt):
            setattr(m, opt, hwinfo.get(opt))

    # create a general section
    if not m.config.has_section("general"):
        m.config.add_section("general")

    # update cfg object
    for opt in ["id", "bootargs", "nickname", "fingerprints", "arch", "ksdata"]:
        if isinstance(getattr(m,opt), list):
            m.config.set("general", opt, " ".join(getattr(m, opt)))
        else:
            m.config.set("general", opt, str(getattr(m, opt)))

    _writemachine(m)

    return True

def add(nickname, hwinfo):
    '''Add the given Machine information to the database'''

    id = 1
    cfg = os.path.join(getoption("machinedb_dir"), str(id), "config")
    while (os.path.isfile(cfg)):
        id = id + 1
        cfg = os.path.join(getoption("machinedb_dir"), str(id), "config")

    hwinfo["id"] = id
    hwinfo["nickname"] = nickname

    return update(id, hwinfo)

def get_ids(dir=None):
    '''Return a list of known machine IDs.'''
    if dir is None:
        dir = getoption("machinedb_dir")
    idlist = list()
    for name in os.listdir(dir):
        (id,ext) = os.path.splitext(name)
        if ext != '.treeinfo':
            continue
        idlist.append(id)
    return idlist

def listmachines(**args):
    '''Return a list of Machines that match all the given args. Recognized args:
    id:     machine id
    arch:   architecture
    nickname: machine name
    fingerprint: one or more machine fingerprints
    ...and basically any other attribute of the Machine object.
    '''
    machinelist = _loadmachines()
    resultlist = list()
    for t in machinelist:
        ok = True
        for (k,v) in args.items():
            if k in ["fingerprint", "fingerprints", "arch"]:
                '''match one or more fingerprints'''
                # coerce as a list
                if isinstance(v,str):
                    v = [v]
                # for each fingerprint ... ensure we have a match
                for f in v:
                    if f not in getattr(t, k):
                        ok = False
                        break
            # attempt a regexp match
            elif k.endswith("~") and re.search(v, getattr(t,k[:-1]), re.I) is None:
                ok = False
                break
            # attempt a regexp match
            elif k.endswith("!") and re.search(v, getattr(t,k[:-1]), re.I) is not None:
                ok = False
                break
            elif (hasattr(t,k) and getattr(t,k) != v):
                ok = False
                break
        if ok:
            resultlist.append(t)
    return resultlist

def dictmachines(args={}):
    '''get a list of Tree objects (in dict format) that match the given args.
    If no args are given, will return a list of all machines.
    Common args:
    arch (e.g. i386, ppc, x86_64...)
    '''
    machinelist = listmachines(**args)
    for t in machinelist:
        del t.config

    return [t.__dict__ for t in machinelist]

    # TODO? return a sorted dictionary
    decorated = [(int(t.id), t.__dict__) for t in machinelist]
    decorated.sort()
    return [dict_ for (key, dict_) in decorated]

def remove(id):
    '''Remove the given machine from the machine database'''
    # This is actually pretty easy.
    cfgdir = os.path.join(getoption("machinedb_dir"), str(id))
    cfg = os.path.join(cfgdir, "config")
    if os.path.isfile(cfg):
        os.unlink(cfg)

        # remove the directory - FIXME (what about other files
        if os.path.isdir(cfgdir):
            os.rmdir(cfgdir)

        logging.debug("Removed machine '%s'" % cfg)
        return True # XMLRPC methods must return a value
    else:
        return False # XMLRPC methods must return a value
